/*!
 *  @file DME.cpp
 *  @mainpage UWB - DME and WLP specifications
 *
 *  @author Daniel Brockmann
 *  @date 13.05.2010
 *
 *  @version 0.1
 *
 *  @brief monitoring the 4 DME message queues (ALME, WLME, MLME, PLME)
 *
 *	There are 4 different message queues necessary for the DME:
 *	\arg ALME : between \c APPLICATION and \c DME
 *	\arg WLME : between \c WLP and \c DME
 *	\arg MLME : between \c MAC and \c DME
 *	\arg PLME : between \c PHY and \c DME
 *	The \c main() -procedure is permanently monitoring all 4 message queues using a simple \c while(1) loop.
 *	The incoming message includes 2 different main informations:
 *	\arg one char-field : the 8 bits for one character stands for 1 special function (see additional table)
 *	\arg one int 		: the integer (32 bit) include the address of the function specific \c structure
 *	There are two different reasons for this kind of transmitted data:
 *	\arg the message queue should obtain as few traffic as possible. in our example: one function and the corresponding parameters
 *	\arg the messages should have a fix length, but the functions have a different quantity of parameters. A address of the structure has always the same length
 *	In the result, each primitiv or message consists of 5 bytes.
 *
 *	In the next step, as soon as a message arrives, the procedure \c *LME_READ() is responsible to decode the incoming function.
 *	Now the necessary function/primitiv is executing in a separate child process. (\c *LME_CHILD )
 *	While the process is executing it is blocked so that is not possible to start it twice. This is necessary to avoid conflicts.
 *
 *	The primitives are defined in a separate \e header-file \c *LME_FUNCTIONS.h.
 *
 *
*/


//! needed standard libraries
#include <iostream>
#include <stdlib.h>

//! function prototypes, important defines
#include "DME.h"

//! responsible for message queue init
#include "MSG_BUFFER.h"

//! defines  main structures like WMIB, DME LIST,...
//#include "STRUCTURES/STRUCTURES.h"


//! defines all the different ALME functions (between application and DME)
#include "DME_FUNCTIONS/ALME_FUNCTIONS.h"
#include "DME_FUNCTIONS/WLME_FUNCTIONS.h"

WMIB WLPInfoBase;


//! a c++ command for input and output
using namespace std;


//! nothing else than monitoring the 4 different message gueues
int main(){

	//! \var buf*LM: the whole structure that is transmitted and received via message queues
	MSGBUF buf_WLME, buf_ALME, buf_MLME, buf_PLME, buf_WLME_P;

	//! \var qid_*LME: include the ID for the message queues; ID is necessary to get access to the queues
	//int qid_LAYER, qid_FKT;
	int qid_WLME, qid_ALME, qid_MLME, qid_PLME, qid_WLME_P, qid_ALME_P, qid_MLME_P, qid_PLME_P;

	//! \var blk_*LME: one list per queue, one field per function, that document whether the function is in use (=1) or not (=0)
	char blk_WLME[WLME_CNT], blk_ALME[ALME_CNT], blk_MLME[MLME_CNT];

	//! init the function - block - lists to 0
	memset(blk_WLME,0,WLME_CNT); memset(blk_ALME,0,ALME_CNT); memset(blk_MLME,0,MLME_CNT);

	//! \var pid_*LME: as soon as a child process is created, the childprocess-ID is saved in this fields
	//! \TODO field should not be necessary
	int pid_WLME[WLME_CNT], pid_ALME[ALME_CNT], pid_MLME[MLME_CNT];

	//! \TODO complete
	int STATUS; pid_t CHILDSTATUS;

	//! init the function - pid - lists to 0
	memset(pid_WLME,0,WLME_CNT); memset(pid_ALME,0,ALME_CNT); memset(pid_MLME,0,MLME_CNT);

	//! create the 4 different message queues
	MESSAGE_QUEUE_INIT(&qid_ALME, &qid_WLME, &qid_MLME, &qid_PLME, &qid_ALME_P, &qid_WLME_P, &qid_MLME_P, &qid_PLME_P);

	//! permanently loop: monitoring the 4 initialised message queues
	while (1){

		//! as soon as a message from application is received it has to be decode (assigned to the specified function) via ALME_READ()
		if(msgrcv(qid_ALME,&buf_ALME, sizeof(MSGBUF), 1,IPC_NOWAIT)!=-1){

			//! print out in console just for information
			printf("Nachr. von APP empfangen: %s\n",buf_ALME.msg);

			//! assign the incoming message to its special primitiv
			ALME_READ(buf_ALME.msg, pid_ALME, blk_ALME);
		}

		//! as soon as a message from WLP is received it has to be decode (assigned to the specified function) via WLME()
		if(msgrcv(qid_WLME,&buf_WLME, sizeof(MSGBUF), 1,IPC_NOWAIT)!=-1){

			//! print out in the console just for information
			printf("Nachr. von WLP empfangen: %s\n",buf_WLME.msg);

			//! assign the incoming message to its special primitiv
			WLME_READ(buf_WLME.msg, pid_WLME, blk_WLME);
		}

//		if(msgrcv(qid_WLME_P,&buf_WLME_P, sizeof(MSGBUF), 1,IPC_NOWAIT)!=-1){
//
//			//! print out in the console just for information
//			printf("Nachr. von xXx empfangen: %s\n",buf_WLME_P.msg);
//		}

		//! as soon as a message from MAC is received it has to be decode (assigned to the specified function) via MLME_READ()
//		if(msgrcv(qid_MLME,&buf_MLME, sizeof(MSGBUF), 1,IPC_NOWAIT)!=-1){

			//! print out in the console just for information
//			printf("Nachr. vom MAC empfangen: %s\n",buf_MLME.msg);

			//! assign the incoming message to its special primitiv
//			MLME_READ((buf_MLME.msg, pid_MLME, blk_MLME);
//		}

		//! as soon as a message from PHY is received it has to be decode (assigned to the specified function) via PLME_READ()
//		if(msgrcv(qid_PLME,&buf_PLME, sizeof(MSGBUF), 1,IPC_NOWAIT)!=-1){

			//! print out in the console just for information
//			printf("Nachr. vom PHY empfangen: %s\n",buf_PLME.msg);

			//! assign the incoming message to its special primitiv
//			PLME_READ((buf_PLME.msg, pid_PLME, blk_PLME);
//		}

		//! delete ZOMBIE (defunct) child-processes that were still in list after finished (instead of WAITPID()
		CHILDSTATUS=waitpid(0, &STATUS, WNOHANG);
	}

}

/*
 * start of the function definitions
 * 1. analyse, decode and assign the messages
 * 2. start a child process for each function
 */

/*!void ALME_READ
 * assign the incoming messages from the application to its
 * corresponding primitiv and starting the assigned childprocess with the correct parameters
 * \param[in] ALME_FKT: number of the primitiv
 * \param[in] pid_t *pid_ALME: pointer to function - process ID - list
 * \param[in] int *blk_ALME: pointer to function - block - list
 * \TODO correct description
 */
void ALME_READ(char ALME_MSG[], pid_t *pid_ALME, char *blk_ALME){

	//! \TODO complete
	int ALME_FKT;

	//! \TODO complete description
	ALME_FKT=ALME_CONVERT((int)ALME_MSG[0]);

	//!switch-case command to separate the different functions and transmit correct parameters to start child process
	switch (ALME_FKT){
		case 0: printf(" ALME_RESET_REQUEST\n"); break;

		case 1: printf(" ALME_WMIB_SET_REQUEST\n"); break;

		case 2: printf(" ALME_WMIB_GET_REQUEST\n");
			//! send WLME_WMIB_GET_request to WLME (WLP)
			ALME_CHILD(ALME_MSG, pid_ALME, blk_ALME, &ALME_WMIB_GET_REQUEST); break;

		case 3: printf(" ALME_DMIB_SET_REQUEST\n"); break;

		case 4: printf(" ALME_DMIB_GET_REQUEST\n"); break;

		case 5: printf(" ALME_WSS_CREATE_REQUEST\n"); break;

		case 6: printf(" ALME_WSS_DESTROY_REQUEST\n"); break;

		case 7: printf(" ALME_WSS_ASSOCIATE_REQUEST\n"); break;

		case 8: printf(" ALME_WSS_DISASSOCIATE_REQUEST\n"); break;

		case 9: printf(" ALME_SCAN_REQUEST\n"); break;

		case 10: printf(" ALME_HIBERNATE_REQUEST\n"); break;

		case 11: printf(" ALME_BEACON_START_REQUEST\n"); break;

		case 12: printf(" ALME_BEACON_STOP_REQUEST\n"); break;

		case 13: printf(" ALME_CHANGE_CHANNEL_REQUEST\n"); break;

		default: printf(" Befehl unbekannt\n"); break;
	}
}

/*!void WLME_READ similar to ALME_READ
 * assign the incoming messages from the WLP to its
 * corresponding primitiv and starting the assigned childprocess with the correct parameters
 * \param[in] WLME_FKT: number of the primitiv
 * \param[in] pid_t *pid_WLME: pointer to function - process ID - list
 * \param[in] int *blk_WLME: pointer to function - block - list
 */
void WLME_READ(char WLME_MSG[], pid_t *pid_WLME, char *blk_WLME){

	//! \TODO complete description
	int WLME_FKT;

	//! \TODO complete description
	WLME_FKT=WLME_CONVERT((int)WLME_MSG[0]);

	//!switch-case command to separate the different functions and transmit correct parameters to start child process
	switch (WLME_FKT){
		case 0: printf(" WLME_RESET_CONFIRM\n");break;
		case 1: printf(" WLME_STREAM_TRACK_INDICATION\n");break;
		case 2: printf(" WLME_WMIB_SET_CONFIRM\n");break;
		case 3: printf(" WLME_WMIB_GET_CONFIRM\n");
			//! release ALME_WMIB_GET_request - BLOCK
			//! send ALME_WMIB_GET_confirm - and provide WMIB to ALME
			WLME_CHILD(WLME_MSG, pid_WLME, blk_WLME, &WLME_WMIB_GET_CONFIRM); break;
		case 4: printf(" WLME_WSS_CREATE_CONFIRM\n");break;
		case 5: printf(" WLME_WSS_DESTROY_CONFIRM\n");break;
		case 6: printf(" WLME_WSS_DESTROY_INDICATION\n");break;
		case 7: printf(" WLME_WSS_ASSOCIATE_CONFIRM\n");break;
		case 8: printf(" WLME_WSS_DISASSOCIATE_CONFIRM\n");break;
		case 9: printf(" WLME_WSS_DISASSOCIATE_INDICATION\n");break;
		default: printf(" Befehl unbekannt\n");break;
	}

}

/*!void MLME_READ similar to ALME_READ
 * assign the incoming messages from the WLP to its
 * corresponding primitiv and starting the assigned childprocess with the correct parameters
 * \param[in] MLME_FKT: number of the primitiv
 * \param[in] pid_t *pid_MLME: pointer to function - process ID - list
 * \param[in] int *blk_MLME: pointer to function - block - list
 */
void MLME_READ(char MLME_MSG[], pid_t *pid_MLME, char *blk_MLME){

	//! \TODO complete description
	int MLME_FKT;

	//! \TODO complete description
	MLME_FKT=MLME_CONVERT((int)MLME_MSG[0]);

	//!switch-case command to separate the different functions and transmit correct parameters to start child process
	switch (MLME_FKT){
		case 0: std::cout << " MLME_RESET\n"; break;
		case 1: std::cout << " MLME_SCAN\n"; break;
		case 2: std::cout << " MLME_SCAN_PLCP_HEADER_RECEIVED\n"; break;
		case 3: std::cout << " MLME_BEACON_START\n"; break;
		case 4: std::cout << " MLME_BEACON_STOP\n"; break;
		case 5: std::cout << " MLME_BEACON_CHANGE_CHANNEL\n"; break;
		case 6: std::cout << " MLME_BEACON\n"; break;
		case 7: std::cout << " MLME_BEACON_MERGE_BP\n"; break;
		case 8: std::cout << " MLME_SET_CAPABILITIES_IE\n"; break;
		case 9: std::cout << " MLME_IDENTIFICATION_IE\n"; break;
		case 10: std::cout << " MLME_PTK\n"; break;
		case 11: std::cout << " MLME_GTK\n"; break;
		case 12: std::cout << " MLME_KEY_UPDATE\n"; break;
		case 13: std::cout << " MLME_SECURITY_VIOLATION\n"; break;
		case 14: std::cout << " MLME_PRF\n"; break;
		case 15: std::cout << " MLME_ASIE_ADD\n"; break;
		case 16: std::cout << " MLME_ASIE_REMOVE\n"; break;
		case 17: std::cout << " MLME_ASIE\n"; break;
		case 18: std::cout << " MLME_MULTICAST_REGISTER\n"; break;
		case 19: std::cout << " MLME_MULTICAST_DEREGISTER\n"; break;
		case 20: std::cout << " MLME_MULTICAST_ACTIVATE\n"; break;
		case 21: std::cout << " MLME_MULTICAST_DEACTIVATE\n"; break;
		case 22: std::cout << " MLME_LINK_EVENT\n"; break;
		case 23: std::cout << " MLME_PROBE\n"; break;
		case 24: std::cout << " MLME_HIBERNATE\n"; break;
		case 25: std::cout << " MLME_WAKE\n"; break;
		case 26: std::cout << " MLME_HIBERNATION_ANCHOR\n"; break;
		case 27: std::cout << " MLME_SLEEP_SCHEDULE\n"; break;
		case 28: std::cout << " MLME_HL_SYNC\n"; break;
		case 29: std::cout << " MLME_RESOURCE\n"; break;
		case 30: std::cout << " MLME_DRP\n"; break;
		case 31: std::cout << " MLME_CONNECTION_CONFIG\n"; break;
		case 32: std::cout << " MLME_RANGE_MEASUREMENT\n"; break;
		case 33: std::cout << " MLME_AS_COMMAND\n"; break;
		default: std::cout << " Befehl unbekannt\n"; break;
	}

}

/*!int ALME_CONVERT
 * \TODO complete description
 */
int ALME_CONVERT(int ALME_CODE){

	switch (ALME_CODE){
		case 0: return 0;
		case 4: return 1;
		case 8: return 2;
		case 12: return 3;
		case 16: return 4;
		case 20: return 5;
		case 24: return 6;
		case 28: return 7;
		case 32: return 8;
		case 36: return 9;
		case 44: return 10;
		case 48: return 11;
		case 52: return 12;
		case 56: return 13;
		default: return -1;
	}

}

/*!int MLME_CONVERT
 * \TODO complete description
 */
int WLME_CONVERT(int WLME_CODE){

	switch (WLME_CODE){
		case 0: return 0;
		case 5: return 1;
		case 8: return 2;
		case 12: return 3;
		case 16: return 4;
		case 20: return 5;
		case 21: return 6;
		case 24: return 7;
		case 29: return 8;
		case 28: return 9;
		default: return -1;
	}

}

/*!int MLME_CONVERT
 * \TODO complete description
 */
int MLME_CONVERT(int MLME_CODE){

	switch (MLME_CODE){
		case 0: return 0;
		case 5: return 1;
		case 8: return 2;
		case 12: return 3;
		case 16: return 4;
		case 20: return 5;
		case 21: return 6;
		case 24: return 7;
		case 29: return 8;
		case 28: return 9;
		default: return -1;
	}

}

/*!void ALME_CHILD
 * checking whether the process/primitiv is already running if not
 * the procedure is creating a child process, sets the blockfield for this function to 1
 * and starting the function in the child process
 * \param[in] int ALME_FKT: number of the primitiv
 * \param[in] pid_t *pid_ALME: pointer to function - process ID - list
 * \param[in] int *blk_ALME: pointer to function - block - list
 * \param[in] int (*ALME_PRIMITIV)(): pointer to a function that should start in the child process
 */
void ALME_CHILD(char ALME_MSG[], pid_t *pid_ALME, char *blk_ALME, int (*ALME_PRIMITIV)(char MSG[])){

	//! \TODO complete
	int ALME_FKT;

	//! \var STATUS_ALME_PRIMITIV get the result 0 from the functions, when successfull done
    int STATUS_ALME_PRIMITIV;

    //! \TOTO complete
    ALME_FKT=ALME_CONVERT((int)ALME_MSG[0]);

    //! if function not already running
    if (BLOCK_CHECK(ALME_FKT,blk_ALME)==0){

    	//! set function - block - field to 1 that means BLOCKED
    	BLOCK_SET(ALME_FKT, blk_ALME);

    	/*! create new childprocess folk() and save its ID in function - pid - field
    	 * \NOTE could be expand by case -1 = ERROR
    	 */
		switch (pid_ALME[ALME_FKT] = fork()) {

			//! all command in case 0 are now executed parallel the rest of the program
			case 0:

				//! start selected function and provide message field
				STATUS_ALME_PRIMITIV=(*ALME_PRIMITIV)(ALME_MSG);

				//! exit childprocess
				exit(0);

			//! all commands in default are executed in regular order
			default:
				break;
		}

    }

    //! if already running print out message to the user
    else {
    	printf("Prozess ist bereits aktiv, bitte erneut versuchen!!\n");
    }

}

/*!void WLME_CHILD equal to ALME_CHILD
 * checking whether the process/primitiv is already running if not
 * the procedure is creating a child process, sets the blockfield for this function to 1
 * and starting the function in the child process
 * \param[in] int WLME_FKT: number of the primitiv
 * \param[in] pid_t *pid_WLME: pointer to function - process ID - list
 * \param[in] int *blk_WLME: pointer to function - block - list
 * \param[in] int (*WLME_PRIMITIV)(): pointer to a function that should start in the child process
 * \TODO complete
 */
void WLME_CHILD(char WLME_MSG[], pid_t *pid_WLME, char *blk_WLME, int (*WLME_PRIMITIV)(char MSG[])){

	//! \TODO complete
	int WLME_FKT;

	//! \var STATUS_WLME_PRIMITIV get the result 0 from the functions, when successfull done
    int STATUS_WLME_PRIMITIV;

    //! \TOTO complete
    WLME_FKT=WLME_CONVERT((int)WLME_MSG[0]);

    //! if function not already running
    if (BLOCK_CHECK(WLME_FKT,blk_WLME)==0){

    	//! set function - block - field to 1 that means BLOCKED
    	BLOCK_SET(WLME_FKT, blk_WLME);

    	/*! create new childprocess folk() and save its ID in function - pid - field
    	 * \NOTE could be expand by case -1 = ERROR
    	 */
		switch (pid_WLME[WLME_FKT] = fork()) {

			//! all command in case 0 are now executed parallel the rest of the program
			case 0:

				//! start selected function and provide message field
				STATUS_WLME_PRIMITIV=(*WLME_PRIMITIV)(WLME_MSG);

				//! exit childprocess
				exit(0);

			//! all commands in default are executed in regular order
			default:
				break;
		}

    }

    //! if already running print out message to the user
    else {
    	printf("Prozess ist bereits aktiv, bitte erneut versuchen!!\n");
    }

}

